import types
from typing import Type, Union

import sapl_base.authorization_subscription_factory
from sapl_base.authorization_subscriptions import AuthorizationSubscription
from sapl_base.constraint_handling.constraint_handler_bundle import ConstraintHandlerBundle
from sapl_base.exceptions import PermissionDenied


class PolicyEnforcementPoint:
    """
    Baseclass for PolicyEnforcementPoints(PEP).
    It handles the enforcement of decorated functions
    """
    constraint_handler_bundle: ConstraintHandlerBundle = None

    def __init__(self, fn: types.FunctionType, *args, **kwargs):
        self._enforced_function = fn
        args_dict = get_named_args_dict(fn, *args, **kwargs)
        self._function_args = args
        self._function_kwargs = kwargs

        try:
            class_object = args_dict.get('self')
            if class_object is None:
                raise KeyError
            self._pos_args = get_class_positional_args(fn, args)
            self.values_dict = {"function": fn, "self": class_object, "args": args_dict}

        except KeyError:
            self._pos_args = get_function_positional_args(fn, args)
            self.values_dict = {"function": fn, "args": args_dict}

    def _get_return_value(self) -> dict:
        """
        Call the decorated function and get the return-value

        :return: The return-value of the decorated function
        """
        self.values_dict["return_value"] = self._enforced_function(**self.values_dict["args"])
        return self.values_dict["return_value"]

    async def _async_get_return_value(self) -> dict:
        """
        Call the decorated function and get the return-value

        :return: The return-value of the decorated function
        """
        self.values_dict["return_value"] = await self._enforced_function(**self.values_dict["args"])
        return self.values_dict["return_value"]

    def _get_subscription(self, subject: Union[str, callable], action: Union[str, callable],
                          resource: Union[str, callable], environment: Union[str, callable], scope: str,
                          enforcement_type: str) -> AuthorizationSubscription:
        """
        Create an AuthorizationSubscription for the decorated function

        :param subject: subject which was provided to the decorator as argument if present
        :param action: action which was provided to the decorator as argument if present
        :param resource: resource which was provided to the decorator as argument if present
        :param environment: environment which was provided to the decorator as argument if present
        :param scope:
        :param enforcement_type: Type of enforcement, with which the function is decorated
        :return: An AuthorizationSubscription
        """
        return sapl_base.authorization_subscription_factory.auth_factory.create_authorization_subscription(self.values_dict, subject, action, resource, environment,
                                                              scope, enforcement_type)

    def _fail_with_bundle(self, exception: Exception) -> None:
        """
        Call all responsible ErrorConstraintHandlerprovider with the given Exception and fail with the Exception of the
        ErrorConstraintHandlerProvider

        :param exception:Exception with which the ErrorConstraintHandlerProvider are called
        """
        try:
            self.constraint_handler_bundle.execute_on_error_handler(exception)
        except Exception as e:
            if isinstance(e, PermissionDenied):
                raise permission_denied_exception
            else:
                raise e

    def _check_if_denied(self, decision) -> None:
        """
        Check if the Decision is DENY and call the method fail_with_bundle with a PermissionDenied Exception if true
        :param decision: Decision
        """
        if decision.decision == "DENY":
            self._fail_with_bundle(permission_denied_exception())


def get_function_positional_args(fn, args):
    """

    :param fn:
    :param args:
    :return:
    """
    return args[0:fn.__code__.co_argcount]


def get_class_positional_args(fn, args):
    """

    :param fn:
    :param args:
    :return:
    """
    return args[1:fn.__code__.co_argcount]


def get_named_args_dict(fn, *args, **kwargs) -> dict:
    """
    Create a dictionary from the args of the function and merge it with the kwargs

    :param fn: function of which args a dict shall be created
    :param args: Arguments provided to the function
    :param kwargs: keyword arguments provided to the function
    """

    if not hasattr(fn, "__code__"):
        args_names = fn.func.__code__.co_varnames[: fn.func.__code__.co_argcount]
    else:
        args_names = fn.__code__.co_varnames[: fn.__code__.co_argcount]

    return {**dict(zip(args_names, args)), **kwargs}


streaming_pep: PolicyEnforcementPoint = None
permission_denied_exception: Type[Exception] = PermissionDenied



import json


class AuthorizationSubscription:
    """
    AuthorizationSubscriptions are sent to the PolicyDecisionPoint and based on their data a Decision is returned.
    """

    def __init__(self, subject, action, resource, environment=None,
                 subscription_id: int = None):
        """
        Create an AuthorizationSubscription Object with the given arguments

        :param subject: Subject, which describes for whom a Decision shall be made. This can be the logged in User for example
        :param action:  For which action shall a Decision be made? This can be a Requesttype like GET or POST for example
        :param resource: The resources for which a Decision shall be made. This can be for example the parameter of a GET request
        :param environment: Optional argument, which describes the environment. This can be for example the current time or location
        :param subscription_id: ID for the object. If no ID is provided an ID is created
        """
        if not (isinstance(subscription_id, int) or subscription_id is None):
            raise TypeError(
                f"subscription_id must be an int, was {subscription_id} type of {subscription_id.__class__}")

        self.subject = subject if subject else {}
        self.action = action if action else {}
        self.resource = resource if resource else {}

        if environment is not None:
            self.environment = environment
        if subscription_id is not None:
            self.subscription_id = subscription_id
        else:
            self.subscription_id = id(self)

    def __repr__(self):
        """
        representative of the object.
        """
        dictionary = self.__dict__.copy()
        representative = ",".join(element + "=" + repr(dictionary.get(element)) for element in dictionary)
        return f"{type(self).__name__}({representative})"

    def __str__(self):
        """
        Sting representation returns this object in json format as a string
        """
        dictionary = self.__dict__.copy()
        dictionary.pop("subscription_id")
        return json.dumps(dictionary, indent=2, skipkeys=True, default=lambda o: str(o))


class MultiSubscription:
    """
    Multiple AuthorizationSubscriptions can be gathered in a MultiSubscription, which can be sent to a
    PolicyDecisionPoint.The PDP will create individuell Decisions for each AuthorizationSubscription, but only one
    request is needed for all Decisions.
    """

    def __init__(
            self, subject=None, action=None, resource=None,
            environment=None,
            authorization_subscriptions=None,
    ):
        if subject is not None:
            self.subject = subject
        if action is not None:
            self.action = action
        if resource is not None:
            self.resource = resource
        if environment is not None:
            self.environment = environment

        self.authorization_subscriptions = authorization_subscriptions

    def __repr__(self):
        """
        representative of the object.
        """
        representative = ",".join(element + "=" + repr(self.__dict__.get(element)) for element in self.__dict__)
        return f"{type(self).__name__}({representative})"

    def __str__(self):
        """
        Sting representation returns this object in json format as a string
        """
        return json.dumps(self.__dict__, indent=2, skipkeys=True, default=lambda o: str(o))



from abc import ABC, abstractmethod
from typing import Any

from sapl_base.decision import Decision


class ConstraintHandlerProvider(ABC):
    """
    Baseclass of a ConstraintHandlerProvider
    """

    @abstractmethod
    def priority(self) -> int:
        """
        ConstraintHandlerProvider are sorted by the value of the priority, when the ConstraintHandlerBundle is created

        :return: value by which ConstraintHandlerProvider are sorted
        """
        return 0

    @abstractmethod
    def is_responsible(self, constraint) -> bool:
        """
        Determine if this ConstraintHandler is responsible for the provided constraint

        :param constraint: A constraint, which can be an Obligation or an Advice, for which the
        ConstraintHandlerProvider checks if it is responsible to handle it.
        :return: Is this ConstraintHandlerProvider responsible to handle the provided constraint
        """
        pass

    @abstractmethod
    def handle(self, argument):
        """
        Abstractmethod, which needs to be implemented by a ConstraintHandlerProvider
        :param argument:
        """
        pass


class ErrorConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider to handle Exceptions,
    which are thrown, when an error occurs while enforcing a decorated function or class
    """

    @abstractmethod
    def handle(self, exception: Exception) -> Exception:
        """
        Abstractmethod, which needs to be implemented by an ErrorConstraintHandlerProvider, which will be called by
        the ConstraintHandlerBundle if the ErrorConstraintHandlerProvider is responsible for a given constraint

        :param exception: Exception which should be handled
        :return: An Exception, which will be provided as Argument
        for the next ErrorConstraintHandlerProvider, or raised after when there is no next
        ErrorConstraintHandlerProvider
        """
        pass


class OnDecisionConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider which can handle a Decision,
    which is received from the PDP
    """

    @abstractmethod
    def handle(self, decision: Decision) -> None:
        """
        Abstractmethod, which needs to be implemented by an OnDecisionHandlerProvider, which will be called by
        the ConstraintHandlerBundle if the OnDecisionHandlerProvider is responsible for a given constraint

        :param decision: Decision which should be handled
        pass
        """


class FunctionArgumentsConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider to handle obligations and advices
    regarding the arguments, which with the enforced function will be called
    """

    @abstractmethod
    def handle(self, function_arguments: dict) -> dict:
        """
        Abstractmethod, which needs to be implemented by an FunctionArgumentsConstraintHandlerProvider, which will be
        called by the ConstraintHandlerBundle if the FunctionArgumentsConstraintHandlerProvider is responsible for a
        given constraint

        :param function_arguments: a dict containing the arguments, with which the enforced function will be called
        :return: a dict, with which the enforced function will be called
        """
        pass


class ResultConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider to handle obligations and advices regarding
    the result of the execution of the enforced function.
    """

    @abstractmethod
    def handle(self, result: Any) -> Any:
        """
        Abstractmethod, which needs to be implemented by an ResultConstraintHandlerProvider, which will be called by
        the ConstraintHandlerBundle if the ResultConstraintHandlerProvider is responsible for a given constraint

        :param result: the result of the enforced function, which was called
        :return: an object, which is treated as if it were the result of the execution of the enforced function
        """
        pass



from sapl_base.policy_enforcement_points.streaming_policy_enforcement_point import StreamingPolicyEnforcementPoint


class BaseStreamingPolicyEnforcementPoint(StreamingPolicyEnforcementPoint):
    async def enforce_till_denied(self, subject, action, resource, environment, scope):
        pass

    async def drop_while_denied(self, subject, action, resource, environment, scope):
        pass

    async def recoverable_if_denied(self, subject, action, resource, environment, scope):
        pass



import json
import logging
import types
from abc import ABC, abstractmethod
from base64 import b64encode
from typing import Coroutine

import aiohttp
import backoff
import requests
from sseclient import SSEClient

from sapl_base.authorization_subscriptions import AuthorizationSubscription
from sapl_base.decision import Decision


class PolicyDecisionPoint(ABC):
    """
    Baseclass of PolicyDecisionPoints(PDP), which creates a PDP based on the current configuration
    Configurations can be provided by a pyproject.toml file

    A PDP returns a Decision, based on the type of PDP, the called method and the
    provided AuthorizationSubscription as argument

    The PDP is a Singleton, which is created on startup can be instantiated directly
    """
    logger: logging.Logger = logging.getLogger(__name__)
    backoff_const_max_time: int
    backoff_expo_max_value: int

    @classmethod
    def from_settings(cls, configuration: dict):
        """
        reads the configuration in the pyproject.toml file and creates a PolicyDecisionPoint(PDP) depending on the
        configuration.
        If no arguments are provided in the pyproject.toml file a RemotePDP is created, with the
        parameters to connect to a SAPL-Server-lt docker container.
        https://github.com/heutelbeck/sapl-policy-engine/tree/master/sapl-server-lt
        """
        logging.basicConfig()
        if configuration.get("dummy", False):
            return DummyPolicyDecisionPoint()
        debug = configuration.get("debug", False)
        base_url = configuration.get("base_url", "http://localhost:8080/api/pdp/")
        key = configuration.get("key", "YJidgyT2mfdkbmL")
        secret = configuration.get("secret", "Fa4zvYQdiwHZVXh")
        verify = configuration.get("verify", False)
        backoff_const_max_time = configuration.get("backoff_const_max_time", 3)
        backoff_expo_max_value = configuration.get("backoff_expo_max_value", 50)
        return RemotePolicyDecisionPoint(base_url, key, secret, verify, debug, backoff_const_max_time,
                                         backoff_expo_max_value)

    @classmethod
    def dummy_pdp(cls):
        """
        Creates a DummyPolicyDecisionPoint without the need to change the configuration

        :return: a DummyPolicyDecisionPoint, which always returns a 'PERMIT'
        """
        return DummyPolicyDecisionPoint()

    @abstractmethod
    async def async_decide(self, subscription: AuthorizationSubscription,
                           pep_decision_stream: types.GeneratorType,
                           decision_events: str = "decide") -> (Decision, Coroutine):
        """
        Request Decisions based on the given AuthorizationSubscription and decision_event
        returns a tuple of the first received Decision and a Coroutine
        which will send received new Decisions to the generator of the calling StreamingPolicyEnforcementpoint

        :param pep_decision_stream: a generator function inside the calling StreamingPolicyEnforcementpoint, to which new Decisions are sent
        :param subscription: AuthorizationSubscription, for which a Decision will be given
        :param decision_events: what kind of decision will be requested from the PDP. defaults to 'decide'
        :return : Tuple of the first received Decision and a Coroutine to send future Decisions to the generator provided as an argument
        """
        pass

    @abstractmethod
    async def async_decide_once(self, subscription: AuthorizationSubscription, decision_events="decide") -> Decision:
        """
        Request only one Decision based on the given AuthorizationSubscription and decision_event

        :param subscription: AuthorizationSubscription, for which a Decision will be given
        :param decision_events: what kind of decision will be requested from the PolicyDecisionPoint. defaults to 'decide'
        :return: A single Decision for the provided AuthorizationSubscription
        """
        pass

    @abstractmethod
    def decide(self, subscription: AuthorizationSubscription, decision_events="decide") -> Decision:
        """
        Blocking method to request a single Decision for the given AuthorizationSubscription and decision_event

        :param subscription: AuthorizationSubscription, for which a Decision will be given
        :param decision_events: what kind of decision will be requested from the PolicyDecisionPoint. defaults to 'decide'
        :return: A single Decision for the provided AuthorizationSubscription
        """
        pass


class DummyPolicyDecisionPoint(PolicyDecisionPoint):
    """
    PolicyDecisionPoint which will always return a PERMIT
    """

    def __init__(self):
        super(DummyPolicyDecisionPoint, self).__init__()
        self.logger.warning(
            "ATTENTION THE APPLICATION USES A DUMMY PDP. ALL AUTHORIZATION REQUEST WILL RESULT IN A SINGLE "
            "PERMIT DECISION. DO NOT USE THIS IN PRODUCTION! THIS IS A PDP FOR TESTING AND DEVELOPING "
            "PURPOSE ONLY!"
        )

    async def async_decide(self, subscription: AuthorizationSubscription, pep_decision_stream: types.GeneratorType,
                           decision_events: str = "decide") -> (Decision, Coroutine):
        """
        implementation of decide, which returns a tuple of a Decision with Permit and a Coroutine which will send a
        Permit to the provided Generator
        """
        return Decision.permit_decision(), self._yield_permit(pep_decision_stream)

    @staticmethod
    async def _yield_permit(pep_decision_stream: types.GeneratorType):
        """
        Send a Permit to the given Generator

        :param pep_decision_stream:  Generator to which the Decision is sent
        """
        pep_decision_stream.send(Decision.permit_decision())

    async def async_decide_once(
            self, subscription: AuthorizationSubscription = None, decision_events: str = None
    ) -> Decision:
        """
        Returns a single Decision with PERMIT

        :return: Decision with Permit
        """
        return Decision.permit_decision()

    def decide(self, subscription: AuthorizationSubscription = None, decision_events=None) -> Decision:
        """
        Returns a single Decision with PERMIT

        :return: Decision with Permit
        """
        return Decision.permit_decision()


async def recreate_stream(details) -> None:
    """
    Function to remove the current Connection to a RemotePDP when an Exception occurs, to establish a new Connection and
    try to reconnect.

    :param details: dictionary of the function which has thrown an Exception
    """
    details['kwargs']['decision_stream'] = None

def set_const_max_time():
    """
    Set the max amount of retrys for a constant backoff from the configuration
    """
    return pdp.backoff_const_max_time


def set_expo_max_value():
    """
    Set the max intervall for an exponential backoff from the configuration
    """
    return pdp.backoff_expo_max_value


class RemotePolicyDecisionPoint(PolicyDecisionPoint, ABC):
    """
    Implementation of a PolicyDecisionPoint(PDP) which connects to an external PDP sends the Authorization and returns
    the Decision received from the external PDP for the given AuthorizationSubscription
    """
    headers = {"Content-Type": "application/json"}

    def __init__(self, base_url: str, key, secret, verify, debug, backoff_const_max_time, backoff_expo_max_value):
        self.debug = debug
        if self.debug:
            self.logger.setLevel('DEBUG')
        if not base_url.endswith('/'):
            self.base_url = base_url + '/'
        else:
            self.base_url = base_url
        self.verify = verify
        if (self.verify is None) or (self.base_url is None):
            raise Exception("No valid configuration for the PDP")
        if key is not None:
            key_and_secret = b64encode(str.encode(f"{key}:{secret}")).decode("ascii")
            self.headers["Authorization"] = f"Basic {key_and_secret}"
        self.logger.info("PDP initialized. \n base url: %s \n key: %s \n secret: %s \n verify: %s", base_url, key,
                         secret, verify)
        self.backoff_const_max_time = backoff_const_max_time
        self.backoff_expo_max_value = backoff_expo_max_value

    @backoff.on_exception(backoff.constant, Exception, max_time=set_const_max_time, raise_on_giveup=False,
                          logger=__name__)
    def decide(self, subscription: AuthorizationSubscription,
               decision_events="decide") -> Decision:
        """
        Blocking method to request a single Decision from the RemotePDP for the given AuthorizationSubscription.

        When an Exception is thrown this method trys to get a Decision again for a maximum 5 seconds.
        On giveup None is returned.

        :param subscription: An Authorization_Subscription for which a Decision is requested
        :param decision_events: For what kind of AuthorizationSubscription should a Decision be returned
        :return: Decision for the given AuthorizationSubscription, or None when no Decision could be evaluated in time.
        """
        if self.debug:
            self.logger.debug("Requesting decision for AuthorizationSubscription: %s", subscription.__str__())
        with requests.post(
                self.base_url + decision_events,
                subscription.__str__(),
                stream=True,
                verify=self.verify,
                headers=self.headers
        ) as stream_response:
            if stream_response.status_code != 200:
                if self.debug:
                    self.logger.debug("Responsecode != 200, was %s . Decision defaults to DENY",
                                      stream_response.status_code)
                return Decision.deny_decision()
            for event in SSEClient(stream_response).events():
                decision = Decision(json.loads(event.data))
                if self.debug:
                    self.logger.debug("Decision : %s", json.dumps(decision))
                return decision

    async def async_decide(self, subscription: AuthorizationSubscription, pep_decision_stream: types.GeneratorType,
                           decision_events: str = "decide") -> (Decision, types.CoroutineType):
        """
        Establish a connection to the RemotePDP and receive new Decisions, which are send to the provided Generator.
        When the connection to the RemotePDP fails, an INDETERMINATE Decision is sent to the Generator and it is
        retried to establish a connection again. Retry works with an exponential backoff and a maximum.

        :param subscription: AuthorizationSubscription for which Decisions should be made
        :param pep_decision_stream: Generator, to which new received Decision are sent
        :param decision_events: For what kind of AuthorizationSubscription should a Decision be returned
        :return: A tuple of the first received Decision and a Coroutine, which will send Decisions to the given pep_decision_stream
        """
        try:
            decision, decision_stream = await self._get_first_decision_and_stream(subscription=subscription,
                                                                                  decision_events=decision_events)
        except Exception as e:
            self.logger.debug("An Error occured while getting the first Decision. Decision defaults to INDETERMINATE")
            decision = {"decision": "INDETERMINATE"}
            decision_stream = None
        return decision, self._update_decision(subscription=subscription, decision_stream=decision_stream,
                                               pep_decision_stream=pep_decision_stream, decision_events=decision_events)

    @backoff.on_exception(backoff.expo, Exception, on_backoff=recreate_stream, max_value=set_expo_max_value)
    async def _update_decision(self, subscription: AuthorizationSubscription, decision_stream: types.AsyncGeneratorType,
                               pep_decision_stream: types.GeneratorType, decision_events: str = "decide") -> None:
        """
        Returns a Coroutine, which will send new Decisions to the provided Generator.
        When an Exception occurs this method sends a INDETERMINATE Decision to the Generator and trys to reestablish a
        connection to the RemotePDP with an exponential backoff

        :param subscription: AuthorizationSubscription for which Decision will be evaluated
        :param decision_stream: Generator, which receives new Decisions from the RemotePDP
        :param pep_decision_stream: Generator, to which new Decisions are sent
        :param decision_events: For what kind of AuthorizationSubscription should a Decision be returned
        """
        if decision_stream is None:
            self.logger.debug("Stream to PDP was cancelled. Retrying to connect to the PDP")
            await pep_decision_stream.send({"decision": "INDETERMINATE"})
            decision_stream = self._get_decision_stream(subscription=subscription, decision_events=decision_events)

        async for decision in decision_stream:
            await pep_decision_stream.send(decision)

    @backoff.on_exception(backoff.constant, Exception, max_time=set_const_max_time)
    async def _get_first_decision_and_stream(self, subscription: AuthorizationSubscription, decision_events: str) -> (
            Decision, types.AsyncGeneratorType):
        """
        Establish a connection to the RemotePDP and return the first Decision together with the Generator,
        which receives new Decisions from the RemotePDP.
        When an Exception occurs this method trys again, until it gives up after 10 seconds.

        :param subscription: AuthorizationSubscription for which Decision will be evaluated
        :param decision_events: For what kind of AuthorizationSubscription should a Decision be returned
        :return: tuple of the first Decision and a Generator which receives new Decisions from the RemotePDP
        """
        decision_stream = self._get_decision_stream(subscription=subscription, decision_events=decision_events)
        decision = await decision_stream.__anext__()
        if decision == {"decision": "INDETERMINATE"}:
            self.logger.debug("First Decision was INDETERMINATE. Defaulting to DENY")
            return Decision.deny_decision(), decision_stream
        return Decision(decision), decision_stream

    async def _get_decision_stream(self, subscription: AuthorizationSubscription,
                                   decision_events: str = "decide") -> types.AsyncGeneratorType:
        """
        Establish a connection to the RemotePDP and yield new Decisions

        :param subscription: AuthorizationSubscription for which Decision will be evaluated
        :param decision_events: For what kind of AuthorizationSubscription should a Decision be returned
        :return: A Generator yielding new Decisions received from the RemotePDP
        """
        async with aiohttp.ClientSession(headers=self.headers, raise_for_status=True) as session:

            async with session.post(self.base_url + decision_events, data=subscription.__str__()
                                    ) as response:

                if response.status != 200:
                    self.logger.debug("Responsecode != 200, was %s . Decision defaults to INDETERMINATE",
                                      response.status)
                    yield {"decision": "INDETERMINATE"}
                else:
                    lines = b''
                    async for line in response.content:
                        lines += line
                        if lines.endswith(b'\n\n'):
                            line_set = lines.splitlines(False)
                            response = ''
                            for item in line_set:
                                response += item.decode('utf-8')
                            data_begin = str.find(response, '{')
                            decision = json.loads(response[data_begin:])
                            self.logger.debug("Decision: %s", response[data_begin:])
                            yield decision
                            lines = b''

    @backoff.on_exception(backoff.constant, Exception, max_time=set_const_max_time, raise_on_giveup=False)
    async def async_decide_once(
            self, subscription: AuthorizationSubscription, decision_events: str = "decide") -> Decision:
        """
        Request a single Decision from the RemotePDP for the given AuthorizationSubscription.

        When an Exception is thrown this method trys again to get a Decision, for a maximum 5 seconds.
        On giveup None is returned.

        :param subscription: An Authorization_Subscription for which a Decision is requested
        :param decision_events: For what kind of AuthorizationSubscription should a Decision be returned
        :return: Decision for the given AuthorizationSubscription, or None when no Decision could be evaluated in time.
        """
        decision_stream = self._get_decision_stream(subscription=subscription, decision_events=decision_events)
        decision = await decision_stream.__anext__()
        await decision_stream.aclose()
        if decision == {"decision": "INDETERMINATE"}:
            self.logger.debug("Decision was INDETERMINATE. Defaulting to DENY")
            return Decision.deny_decision()
        return Decision(decision)


pdp: PolicyDecisionPoint



from sapl_base.authorization_subscription_factory import AuthorizationSubscriptionFactory
from sapl_base.authorization_subscriptions import AuthorizationSubscription


class BaseAuthorizationSubscriptionFactory(AuthorizationSubscriptionFactory):
    """
    Basic implementation of an AuthorizationSubscriptionFactory, which is used, when the used framework is not set.
    """

    def create_authorization_subscription(self, values: dict, subject, action, resource, environment, scope,
                                          enforcement_type):
        pass

    def _default_subject_function(self, values: dict) -> dict:
        pass

    def _default_action_function(self, values: dict) -> dict:
        pass

    def _default_resource_function(self, values: dict) -> dict:
        pass

    def _add_contextvar_to_values(self, values: dict):
        pass

    def _identify_type(self, values: dict):
        pass

    def _valid_combination(self, fn_type, enforcement_type):
        pass

    def _create_subscription_for_type(self, fn_type, values: dict, subject, action, resource, environment,
                                      scope) -> AuthorizationSubscription:
        pass



import asgiref.sync

from sapl_base.constraint_handling.constraint_handler_service import constraint_handler_service
from sapl_base.decision import Decision
import sapl_base.policy_decision_points
#from sapl_base.policy_decision_points import pdp
from sapl_base.policy_enforcement_points.policy_enforcement_point import PolicyEnforcementPoint


class AsyncPolicyEnforcementPoint(PolicyEnforcementPoint):
    """
    Implementation of an Asynchron PolicyEnforcementPoint
    """

    async def post_enforce(self, subject, action, resource, environment, scope):
        """
        Execute the decorated function, before creating an AuthorizationSubscription and receiving a Decision.
        The return-value of the executed function is sent to the PDP as a resource
        :param subject: subject which was provided to the decorator as argument if present
        :param action: action which was provided to the decorator as argument if present
        :param resource: resource which was provided to the decorator as argument if present
        :param environment: environment which was provided to the decorator as argument if present
        :param scope:

        :return: The return-value of the decorated function after it was enforced

        """
        await self._async_get_return_value()
        return await self._post_enforce_handling(subject, action, resource, environment, scope)

    async def pre_enforce(self, subject, action, resource, environment, scope):
        """
        Create a AuthorizationSubscription for the decorated function and request a Decision, before the function is executed.

        :param subject: subject which was provided to the decorator as argument if present
        :param action: action which was provided to the decorator as argument if present
        :param resource: resource which was provided to the decorator as argument if present
        :param environment: environment which was provided to the decorator as argument if present
        :param scope:
        :return: The return-value of the decorated function after it was enforced
        """
        subscription = await asgiref.sync.sync_to_async(self._get_subscription)(subject, action, resource, environment,
                                                                                scope, "pre_enforce")
        decision = await sapl_base.policy_decision_points.pdp.async_decide_once(subscription)
        if decision is None:
            decision = Decision.deny_decision()
        self.constraint_handler_bundle = constraint_handler_service.build_pre_enforce_bundle(decision)
        self._check_if_denied(decision)
        self.constraint_handler_bundle.execute_on_decision_handler(decision)
        self.constraint_handler_bundle.execute_function_arguments_mapper(self.values_dict["args"])
        return_value = await self._async_get_return_value()
        return self.constraint_handler_bundle.execute_result_handler(return_value)

    async def pre_and_post_enforce(self, subject, action, resource, environment, scope):
        """
        Combination of pre_enforce and post_enforce.
        The decorated function is enforced before it is executed and after it is executed.

        :param subject: subject which was provided to the decorator as argument if present
        :param action: action which was provided to the decorator as argument if present
        :param resource: resource which was provided to the decorator as argument if present
        :param environment: environment which was provided to the decorator as argument if present
        :param scope:
        :return: The return-value of the decorated function after it was enforced
        """
        self.values_dict["return_value"] = await self.pre_enforce(subject, action, resource, environment, scope)
        return await self._post_enforce_handling(subject, action, resource, environment, scope)

    async def _post_enforce_handling(self, subject, action, resource, environment, scope):
        """
        Gets called after the decorated function is executed to enforce the function

        :param subject: subject which was provided to the decorator as argument if present
        :param action: action which was provided to the decorator as argument if present
        :param resource: resource which was provided to the decorator as argument if present
        :param environment: environment which was provided to the decorator as argument if present
        :param scope:
        :return: The return-value of the decorated function after it was enforced
        """
        subscription = await asgiref.sync.sync_to_async(self._get_subscription)(subject, action, resource, environment,
                                                                                scope, "post_enforce")
        decision = await sapl_base.policy_decision_points.pdp.async_decide_once(subscription)
        if decision is None:
            decision = Decision.deny_decision()
        self.constraint_handler_bundle = constraint_handler_service.build_post_enforce_bundle(decision)
        self._check_if_denied(decision)
        self.constraint_handler_bundle.execute_on_decision_handler(decision)
        return self.constraint_handler_bundle.execute_result_handler(self.values_dict["return_value"])



import contextvars
from abc import abstractmethod, ABC
from typing import Callable

from .authorization_subscriptions import AuthorizationSubscription, MultiSubscription

client_request = contextvars.ContextVar('request')


class AuthorizationSubscriptionFactory(ABC):
    """
    Baseclass of an AuthorizationSubscriptionFactory, which can be inherited to create a framework specific
    AuthorizationSubscriptionFactory.
    """
    subject_functions: list[Callable[[dict], dict]] = None

    def init_factory(self, subject_functions: list[[Callable[[dict], dict]]] = None):
        self.subject_functions = subject_functions

    def _create_subscription(self, values: dict, subject=None, action=None, resource=None,
                             environment=None) -> AuthorizationSubscription:
        """
        Create an AuthorizationSubscription for the decorated function, with the arguments provided to the decorator

       :param values: Dictionary which contains data related to the decorated function
       :param subject: subject with which the function was decorated. None if not specified
       :param action:  action with which the function was decorated. None if not specified
       :param resource: resource with which the function was decorated. None if not specified
       :param environment: environment with which the function was decorated. None if not specified
       :return: An authorization_subscription which can be sent to a pdp to get an authorization_decision
        """

        if subject is not None:
            _subject = self._argument_is_callable(subject, values)
        else:
            _subject = self._default_subject_function(values)
            if self.subject_functions:
                for fun in self.subject_functions:
                    _subject.update(fun(values))
            if not _subject:
                _subject = "anonymous"

        if action is not None:
            _action = self._argument_is_callable(action, values)
        else:
            _action = self._default_action_function(values)

        if resource is not None:
            _resource = self._argument_is_callable(resource, values)
        else:
            _resource = self._default_resource_function(values)

        if environment is not None:
            _environment = self._argument_is_callable(environment, values)
        else:
            _environment = {}

        return AuthorizationSubscription(self._remove_empty_dicts(_subject), self._remove_empty_dicts(_action),
                                         self._remove_empty_dicts(_resource),
                                         self._remove_empty_dicts(_environment))

    @staticmethod
    def _argument_is_callable(argument, values: dict):
        """
        Checks if the given argument is a callable and calls it with the given dictionary of values as argument

        :param argument: given argument which was provided to the SAPL Decorator as argument for certain parameter
        :param values: dictionary of the values which are used to create the authorization_subscription
        :return: given argument if it is not a callable, otherwise the return-value of the argument called with values
        """
        if callable(argument):
            return argument(values)
        else:
            return argument

    @abstractmethod
    def _default_subject_function(self, values: dict) -> dict:
        """
        Default method which is called to create the subject of the AuthorizationSubscription, if no function is
        provided as argument to the decorator of a decorated function

        :param values: dictionary containing all values from which a subject can be created.
        :return: A dictionary which will be provided as subject, when an AuthorizationSubscription is created
        """
        pass

    @abstractmethod
    def _default_action_function(self, values: dict) -> dict:
        """
        Default method which is called to create the action of the AuthorizationSubscription, if no function is
        provided as argument to the decorator of a decorated function

        :param values: dictionary containing all values from which an action can be created.
        :return: A dictionary which will be provided as action, when an AuthorizationSubscription is created
        """
        pass

    @abstractmethod
    def _default_resource_function(self, values: dict) -> dict:
        """
        Default method which is called to create the resource of the AuthorizationSubscription, if no function is
        provided as argument to the decorator of a decorated function

        :param values: dictionary containing all values from which a resource can be created.
        :return: A dictionary which will be provided as resource, when an AuthorizationSubscription is created
        """
        pass

    @abstractmethod
    def create_authorization_subscription(self, values: dict, subject, action, resource,
                                          environment, scope, enforcement_type):
        """
        Create an AuthorizationSubscription with the given dictionary and arguments

        The returned AuthorizationSubscription is dependent of the framework and the decorated function

        :param enforcement_type: the type of enforcement, with which the function is decorated
        :param scope: Argument which creates a AuthorizationSubscription according to the given scope instead of evaluating the scope based on other parameter
        :param values: Dictionary which contains data related to the decorated function (class if present, function and dict with named args )
        :param subject: subject with which the function was decorated. None if not specified
        :param action:  action with which the function was decorated. None if not specified
        :param resource: resource with which the function was decorated. None if not specified
        :param environment: environment with which the function was decorated. None if not specified
        :return: An authorization_subscription which can be sent to a pdp to get an authorization_decision
        """
        pass

    def _remove_empty_dicts(self, dictionary: dict):
        """
        A helper function to prevent empty dictionary's from being added to the authorization_subscription

        :param dictionary: A dictionary which will be stripped of it's empty dictionary's
        :return: The given dictionary stripped of empty dictionary's inside itself, or None if the given dictionary is empty
        """
        dict_copy = dictionary.copy()
        for k, v in dictionary.items():
            if isinstance(v, dict):
                dict_copy[k] = self._remove_empty_dicts(v)

            if dict_copy[k] is None:
                dict_copy.pop(k)

        if not dict_copy:
            return None

        return dict_copy


class MultiSubscriptionBuilder:
    """
    Can be used to create a Multi subscription by adding single Authorization_subscription
    """
    SUBJECT_ID = "subjectID"
    ACTION_ID = "actionID"
    RESOURCE_ID = "resourceID"
    ENVIRONMENT_ID = "environmentID"
    AUTHORIZATION_SUBSCRIPTION_ID = "authorization_subscriptionID"

    def __init__(self):
        self._built = False
        self.subject = []
        self.action = []
        self.resource = []
        self.environment = []
        self.authorization_subscription = []

    def with_authorization_subscription(self, authorization_subscription: AuthorizationSubscription):
        """
        Adds the given AuthorizationSubscription to the MultiSubscriptionBuilder
        :param authorization_subscription: AuthorizationSubscription, which will be added to the MultiSubscriptionBuilder
        """
        dic = dict()
        self._add_subject(authorization_subscription, dic)
        self._add_action(authorization_subscription, dic)
        self._add_resource(authorization_subscription, dic)
        self._add_environment(authorization_subscription, dic)
        self.authorization_subscription.append({authorization_subscription.subscription_id: dic})

    def _add_subject(self, authorization_subscription: AuthorizationSubscription, dictionary: dict):
        if authorization_subscription.subject:
            self.subject.append(authorization_subscription.subject)
            dictionary[self.SUBJECT_ID] = len(self.subject) - 1

    def _add_action(self, authorization_subscription: AuthorizationSubscription, dictionary: dict):
        if authorization_subscription.action:
            self.action.append(authorization_subscription.action)
            dictionary[self.ACTION_ID] = len(self.action) - 1

    def _add_resource(self, authorization_subscription: AuthorizationSubscription, dictionary: dict):
        if authorization_subscription.resource:
            self.resource.append(authorization_subscription.resource)
            dictionary[self.RESOURCE_ID] = len(self.resource) - 1

    def _add_environment(self, authorization_subscription: AuthorizationSubscription, dictionary: dict):
        if authorization_subscription.environment:
            self.environment.append(authorization_subscription.environment)
            dictionary[self.ENVIRONMENT_ID] = len(self.environment) - 1

    @staticmethod
    def _remove_empty_list(subscription_list):
        """
        Helper function to remove empty lists

        :param subscription_list: A list which contains arguments, which will
        be passed to a Multi subscription as arguments
        :return: None if the list is empty, otherwise the given list.
        """
        if not subscription_list:
            return None
        return subscription_list

    def build(self):
        """
        :return: A multi subscription created from the given Authorization_subscriptions
        """
        if self._built:
            raise Exception("already built")
        self._built = True
        return MultiSubscription(self._remove_empty_list(self.subject), self._remove_empty_list(self.action),
                                 self._remove_empty_list(self.resource), self._remove_empty_list(self.environment),
                                 self._remove_empty_list(self.authorization_subscription))


auth_factory = None
